/*
 *******************************************************
  
 cat an mice benchmark, wodes08, libfaudes 2.09

 this file provides models for the individual cat k and mouse k
 residing on n levels of the maze hierarchy, as well as 
 specifications for the cats and mice
  
 *******************************************************
 */ 
 
#include <stdio.h>
#include <iostream>
#include "libfaudes.h"
#include "catsMice_individual_models.h"
#include <vector>


//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%% Create cat k (residing on n maze levels) %%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// Cat(rCat, n, k)
void Cat(cGenerator& rCat, const int n, const int k)
{
    FD_DF("Cat(" << rCat.Name() << "," << n << "," << k <<")");
    // helper variable to hold current event
    Idx event; 
      
    // Prepare strings "k" for constructing the statenames of cat number k
      std::string kstr=ToStringInteger(k);
      // prepare the cat generator
      rCat.Clear();
      // set name
      rCat.Name("Cat"+kstr);
      
      // catch invalid inputs
      if(n<1)
      {
          FD_DF("Cat:: the minimum value for n is 1: empty result.");
          return;
      }
    
      if(k<0)
      {
          FD_DF("Cat::  the minimum value for k is 1: empty result.");
          return;
      }
      // vector for the cat states on the first level. There is one state for each of the 5 rooms
      std::vector<Idx> fiveStates(5);
      for(int i = 0; i< 5; i++){
          Idx newState = rCat.InsState("C" + ToStringInteger(i) + "("+ToStringInteger(1)+","+kstr+")");
          fiveStates[i] = newState;
      }
    
      // The state that connects the first maze level to the second maze level 
      // (if there is a second level) corresponds to room 0 
      int connectedRoom = 0; // the index of the room
      Idx connectedState = fiveStates[0]; // the index of the state on maze level 1
      //////////////////////////
      // construction of level 1
      //////////////////////////
      
      // room 0 is the initial room for the cat in maze level 1
      rCat.SetInitState(fiveStates[0]);
      rCat.SetMarkedState(fiveStates[0]);
      
      // the events that describe transitions of the cat between the rooms on maze level 1
      // are created. The label "cx-y(n,k)" denotes a transition of cat k from room x to room y on maze level n
      event = rCat.InsControllableEvent("c1-2("+ToStringInteger(1)+","+kstr+")");
      rCat.SetTransition(fiveStates[1],event,fiveStates[2]);
      
      event = rCat.InsUncontrollableEvent("c1-3("+ToStringInteger(1)+","+kstr+")");
      rCat.SetTransition(fiveStates[1],event,fiveStates[3]);
    
      event = rCat.InsControllableEvent("c2-0("+ToStringInteger(1)+","+kstr+")");
      rCat.SetTransition(fiveStates[2],event,fiveStates[0]);
      
      event = rCat.InsControllableEvent("c0-1("+ToStringInteger(1)+","+kstr+")");
      rCat.SetTransition(fiveStates[0],event,fiveStates[1]);
      
      event = rCat.InsControllableEvent("c0-3("+ToStringInteger(1)+","+kstr+")");
      rCat.SetTransition(fiveStates[0],event,fiveStates[3]);
      
      event = rCat.InsUncontrollableEvent("c3-1("+ToStringInteger(1)+","+kstr+")");
      rCat.SetTransition(fiveStates[3],event,fiveStates[1]);
      
      event = rCat.InsControllableEvent("c3-4("+ToStringInteger(1)+","+kstr+")");
      rCat.SetTransition(fiveStates[3],event,fiveStates[4]);
      
      event = rCat.InsControllableEvent("c4-0("+ToStringInteger(1)+","+kstr+")");
      rCat.SetTransition(fiveStates[4],event,fiveStates[0]);  
      
      //////////////////////////////////////////////////////
      // construction of higher levels starting from level 2
      //////////////////////////////////////////////////////
      
      // all remaining levels from 2 to n are added to the cat model
      for(int l = 2; l <= n; l++){
          // insert the 5 states for each room of the current level
          for(int i = 0; i < 5; i++){
              Idx newState = rCat.InsState("C" + ToStringInteger(i) + "("+ToStringInteger(l)+","+kstr+")");
              fiveStates[i] = newState;
          }
          // insert the events for the current level of the maze
          event = rCat.InsControllableEvent("c1-2("+ToStringInteger(l)+","+kstr+")");
          rCat.SetTransition(fiveStates[1],event,fiveStates[2]);
          
          event = rCat.InsUncontrollableEvent("c1-3("+ToStringInteger(l)+","+kstr+")");
          rCat.SetTransition(fiveStates[1],event,fiveStates[3]);
        
          event = rCat.InsControllableEvent("c2-0("+ToStringInteger(l)+","+kstr+")");
          rCat.SetTransition(fiveStates[2],event,fiveStates[0]);
          
          event = rCat.InsControllableEvent("c0-1("+ToStringInteger(l)+","+kstr+")");
          rCat.SetTransition(fiveStates[0],event,fiveStates[1]);
          
          event = rCat.InsControllableEvent("c0-3("+ToStringInteger(l)+","+kstr+")");
          rCat.SetTransition(fiveStates[0],event,fiveStates[3]);
          
          event = rCat.InsUncontrollableEvent("c3-1("+ToStringInteger(l)+","+kstr+")");
          rCat.SetTransition(fiveStates[3],event,fiveStates[1]);
          
          event = rCat.InsControllableEvent("c3-4("+ToStringInteger(l)+","+kstr+")");
          rCat.SetTransition(fiveStates[3],event,fiveStates[4]);
          
          event = rCat.InsControllableEvent("c4-0("+ToStringInteger(l)+","+kstr+")");
          rCat.SetTransition(fiveStates[4],event,fiveStates[0]);  
          
          // establish connection to lower maze level: The connected room is stored in the variable "connectedRoom". 
          // Hence, the connected state on the current maze level can be determined as the state for 
          // "connectedRoom" in the vector fiveStates. The connected state on the previous level is already given 
          // by the variable "connectedState". The introduced events cx-y(k) express a transition from level
          // x to level y by cat k.  
          event = rCat.InsControllableEvent("c"+ToStringInteger(l)+"-"+ToStringInteger(l-1)+"("+kstr+")");
          rCat.SetTransition(fiveStates[connectedRoom],event,connectedState);  
          
          event = rCat.InsControllableEvent("c"+ToStringInteger(l-1)+"-"+ToStringInteger(l)+"("+kstr+")");
          rCat.SetTransition(connectedState,event,fiveStates[connectedRoom]);  
          
          // update the connection variables
          connectedRoom = (connectedRoom + 1) % 5;
          connectedState = fiveStates[connectedRoom];
      }// for(int l = 2; l <= n; l++)
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%% Create mouse k (residing on n maze levels) %%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// Mouse( rMouse, n, k)
void Mouse(cGenerator& rMouse, const int n, const int k)
{
    FD_DF("Mouse(" << rMouse.Name() << "," << n "," << k << ")");
  
    // helper variable to hold current event
	Idx event; 
  
	  // Prepare strings "k" for constructing the statenames of cat number k
      std::string kstr=ToStringInteger(k);
      // prepare mouse generator
      rMouse.Clear();
      // set name
      rMouse.Name("Mouse"+kstr);
      
      // catch invalid inputs
      if(n<1)
      {
    	 FD_DF("Mouse:: the minimum value for n is 1: empty result.");
    	 return;
      }
    
      if(k<0)
      {
    	  FD_DF("Mouse:: the minimum value for k is 1: empty result.");
    	  return;
      }
      // vector for the mouse states on the n-th level. There is one state for each of the 5 rooms
      std::vector<Idx> fiveStates(5);
      for(int i = 0; i< 5; i++){
          Idx newState = rMouse.InsState("M" + ToStringInteger(i) + "("+ToStringInteger(n)+","+kstr+")");
          fiveStates[i] = newState;
      }
    
      // The state that connects the n-th maze level to the n-1st maze level 
      // (if there is such level) corresponds to room (n-2) % 5 according to the benchmark description 
      int connectedRoom = (n - 2) % 5; // index of the connected room
      Idx connectedState = fiveStates[connectedRoom]; // index of the connected state on maze level n
      
      /////////////////////////////////////////////
      // Construction of the highest maze level (n)
      /////////////////////////////////////////////
      
      // room 4 is the initial room for the mouse in maze level n
      rMouse.SetInitState(fiveStates[4]);
      rMouse.SetMarkedState(fiveStates[4]);
      
      // the events that describe transitions of the mouse between the rooms on maze level n
      // are created. The label "mx-y(n,k)" denotes a transition of mouse k from room x to room y on maze level n
      event = rMouse.InsControllableEvent("m0-2("+ToStringInteger(n)+","+kstr+")");
      rMouse.SetTransition(fiveStates[0],event,fiveStates[2]);
      
      event = rMouse.InsControllableEvent("m0-4("+ToStringInteger(n)+","+kstr+")");
      rMouse.SetTransition(fiveStates[0],event,fiveStates[4]);
    
      event = rMouse.InsControllableEvent("m1-0("+ToStringInteger(n)+","+kstr+")");
      rMouse.SetTransition(fiveStates[1],event,fiveStates[0]);
      
      event = rMouse.InsControllableEvent("m2-1("+ToStringInteger(n)+","+kstr+")");
      rMouse.SetTransition(fiveStates[2],event,fiveStates[1]);
      
      event = rMouse.InsControllableEvent("m3-0("+ToStringInteger(n)+","+kstr+")");
      rMouse.SetTransition(fiveStates[3],event,fiveStates[0]);
      
      event = rMouse.InsControllableEvent("m4-3("+ToStringInteger(n)+","+kstr+")");
      rMouse.SetTransition(fiveStates[4],event,fiveStates[3]);
      
      //////////////////////////////////////////////////////
      // Construction of the lower maze levels from n-1 to 1
      //////////////////////////////////////////////////////
      
      // all remaining levels from n-1 to 1 are added to the mouse model
      for(int l = n-1; l > 0; l--){
          // insert the 5 states for each room of the current level
          for(int i = 0; i< 5; i++){
              Idx newState = rMouse.InsState("M" + ToStringInteger(i) + "("+ToStringInteger(l)+","+kstr+")");
              fiveStates[i] = newState;
          }
          // insert the events for the current level of the maze
          event = rMouse.InsControllableEvent("m0-2("+ToStringInteger(l)+","+kstr+")");
          rMouse.SetTransition(fiveStates[0],event,fiveStates[2]);
    
          event = rMouse.InsControllableEvent("m0-4("+ToStringInteger(l)+","+kstr+")");
          rMouse.SetTransition(fiveStates[0],event,fiveStates[4]);
    
          event = rMouse.InsControllableEvent("m1-0("+ToStringInteger(l)+","+kstr+")");
          rMouse.SetTransition(fiveStates[1],event,fiveStates[0]);
    
          event = rMouse.InsControllableEvent("m2-1("+ToStringInteger(l)+","+kstr+")");
          rMouse.SetTransition(fiveStates[2],event,fiveStates[1]);
    
          event = rMouse.InsControllableEvent("m3-0("+ToStringInteger(l)+","+kstr+")");
          rMouse.SetTransition(fiveStates[3],event,fiveStates[0]);
    
          event = rMouse.InsControllableEvent("m4-3("+ToStringInteger(l)+","+kstr+")");
          rMouse.SetTransition(fiveStates[4],event,fiveStates[3]);

          // establish connection to upper maze level: The connected room is stored in the variable "connectedRoom". 
          // Hence, the connected state on the current maze level can be determined as the state for 
          // "connectedRoom" in the vector fiveStates. The connected state on the previous level is already given 
          // by the variable "connectedState". The introduced events mx-y(k) express a transition from level
          // x to level y by mouse k.  
          event = rMouse.InsControllableEvent("m"+ToStringInteger(l)+"-"+ToStringInteger(l+1)+"("+kstr+")" );
          rMouse.SetTransition(fiveStates[connectedRoom],event,connectedState);
    
          event = rMouse.InsControllableEvent("m"+ToStringInteger(l+1)+"-"+ToStringInteger(l)+"("+kstr+")" );
          rMouse.SetTransition(connectedState,event,fiveStates[connectedRoom]);
          // update connected state for next level
          connectedRoom = (connectedRoom + 5 - 1) % 5;
          connectedState = fiveStates[connectedRoom];
      }// for(int l = n-1; l > 0; l--)
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%% Create specification of cat k (residing on n maze levels %%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// CatSpecification(rCat, rCatSpec, catNumber, n, k)
void CatSpecification(cGenerator& rCat, cGenerator& rCatSpec, const int catNumber, const int n, const int k){
    FD_DF("CatSpecification(" << rCat.Name() << "," << rCatSpec.Name() << "," << catNumber << "," << n << "," << k << ")");
    Idx absentState, presentState, event;
    std::string kstr = ToStringInteger(k);
    // vectors for all specifications in all rooms in all levels: these specification state that the cat rCat
    // must never be in the same room with any of the mice
    std::vector<cGenerator> catsRooms0(n);
    cGenerator catRoom0;
    std::vector<cGenerator> catsRooms1(n);
    cGenerator catRoom1;
    std::vector<cGenerator> catsRooms2(n);
    cGenerator catRoom2;
    std::vector<cGenerator> catsRooms3(n);
    cGenerator catRoom3;
    std::vector<cGenerator> catsRooms4(n);
    cGenerator catRoom4;
    
    // preparation of the specification generator
    rCatSpec.Clear();
    cGenerator tmp;    
    // the cat specification is initialized with the corresponding cat model
    Cat(rCatSpec,n,catNumber); 
    rCat = rCatSpec;
    std::string catstr = ToStringInteger(catNumber);
    
    // go through all rooms and levels, and determine the restrictions for the cat catNumber
    for(int l = 1; l <= n; l++){  
        std::string lstr = ToStringInteger(l);
        catRoom0.Clear();
        // ============================================
        // Restrictions in ROOM 0 for the current level
        // ============================================
        // for  cat catNumber, mutual exclusion with all mice in room 0: whenever the cat is absent, any 
        // mouse can be in the room, if the cat is present, no mouse may enter the room
        absentState = catRoom0.InsState("!SC0("+lstr+","+catstr+")" );
        presentState = catRoom0.InsState("SC0("+lstr+","+catstr+")" );
        // in level 1, the cat is initially present in room 1
        if(l == 1){
            catRoom0.SetInitState(presentState);
            catRoom0.SetMarkedState(presentState);  
        }
        // in all levels different from 1, the cat is initially absent in room 1
        else{
            catRoom0.SetInitState(absentState);
            catRoom0.SetMarkedState(absentState);               
        }
        // insert all mouse events that are allowed in the absent state of the cat
        for(int j = 1; j<=k; j++){
            event = catRoom0.InsControllableEvent("m1-0("+lstr+","+ToStringInteger(j)+")" );
            catRoom0.SetTransition(absentState,event,absentState);
            
            event = catRoom0.InsControllableEvent("m3-0("+lstr+","+ToStringInteger(j)+")" );
            catRoom0.SetTransition(absentState,event,absentState);
            // consider the mice that can come from one level up in level 1,6,...
            if( ( (l % 5) == 1) && n > l ){
                event = catRoom0.InsControllableEvent("m"+ToStringInteger(l+1)+"-"+lstr+"("+ToStringInteger(j)+")" );
                catRoom0.SetTransition(absentState,event,absentState);                
            }
            // consider the mice that can come from one level down in level 2,7,...
            if( (l % 5) == 2){
                event = catRoom0.InsControllableEvent("m"+ToStringInteger(l-1)+"-"+lstr+"("+ToStringInteger(j)+")" );
                catRoom0.SetTransition(absentState,event,absentState);                   
            }
        }
        
        // insert transitions between the absent state and the present state of the cat in room 1
        event = catRoom0.InsControllableEvent("c2-0("+lstr+","+catstr+")" );
        catRoom0.SetTransition(absentState,event,presentState);
        
        event = catRoom0.InsControllableEvent("c4-0("+lstr+","+catstr+")" );
        catRoom0.SetTransition(absentState,event,presentState);
        
        event = catRoom0.InsControllableEvent("c0-1("+lstr+","+catstr+")" );
        catRoom0.SetTransition(presentState,event,absentState);
        
        event = catRoom0.InsControllableEvent("c0-3("+lstr+","+catstr+")" );
        catRoom0.SetTransition(presentState,event,absentState);
        
        // if level 1,6,... and the next higher level exists, consider
        // transitions from and to higher level 
        if( ( (l % 5) == 1) && n > l ){
            event = catRoom0.InsControllableEvent("c"+ToStringInteger(l+1)+"-"+lstr+"("+catstr+")" );
            catRoom0.SetTransition(absentState,event,presentState);
            
            event = catRoom0.InsControllableEvent("c"+lstr+"-"+ToStringInteger(l+1)+"("+catstr+")" );
            catRoom0.SetTransition(presentState,event,absentState);                                         
        }
        // if level 2,7,... and , consider
        // transitions from and to lower level 
        if( (l % 5) == 2){
            event = catRoom0.InsControllableEvent("c"+ToStringInteger(l-1)+"-"+lstr+"("+catstr+")" );
            catRoom0.SetTransition(absentState,event,presentState);
            
            event = catRoom0.InsControllableEvent("c"+lstr+"-"+ToStringInteger(l-1)+"("+catstr+")" );
            catRoom0.SetTransition(presentState,event,absentState);                 
        }
        
        // compose the current specification with the already computed spec
        tmp = rCatSpec;
        cParallel(tmp,catRoom0,rCatSpec);

        // ===========================================
        // Restrictions in ROOM 1 of the current level
        // ===========================================
        // the same considerations as for ROOM 0 are made:
        // for one cat, mutual exclusion with all mice in room 1
        catRoom1.Clear();
        absentState = catRoom1.InsState("!SC1("+lstr+","+catstr+")" );
        presentState = catRoom1.InsState("SC1("+lstr+","+catstr+")" );
        
        catRoom1.SetInitState(absentState);
        catRoom1.SetMarkedState(absentState);  
        for(int j = 1; j<=k; j++){
            event = catRoom1.InsControllableEvent("m2-1("+lstr+","+ToStringInteger(j)+")" );
            catRoom1.SetTransition(absentState,event,absentState);
            // consider the mice that can come from one level up in level 2,7,...
            if( ( (l % 5) == 2) && n > l ){
                event = catRoom1.InsControllableEvent("m"+ToStringInteger(l+1)+"-"+lstr+"("+ToStringInteger(j)+")" );
                catRoom1.SetTransition(absentState,event,absentState);                
            }
            // consider the mice that can come from one level down in level 3,8,...
            if( (l % 5) == 3){
                event = catRoom1.InsControllableEvent("m"+ToStringInteger(l-1)+"-"+lstr+"("+ToStringInteger(j)+")" );
                catRoom1.SetTransition(absentState,event,absentState);                   
            }
        }
        // insert the transitions between the absent and the present state of the cat in room 1
        event = catRoom1.InsControllableEvent("c0-1("+lstr+","+catstr+")" );
        catRoom1.SetTransition(absentState,event,presentState);
        
        event = catRoom1.InsUncontrollableEvent("c3-1("+lstr+","+catstr+")" );
        catRoom1.SetTransition(absentState,event,presentState);
        
        event = catRoom1.InsControllableEvent("c1-2("+lstr+","+catstr+")" );
        catRoom1.SetTransition(presentState,event,absentState);
        
        event = catRoom1.InsUncontrollableEvent("c1-3("+lstr+","+catstr+")" );
        catRoom1.SetTransition(presentState,event,absentState);
        
        // if level 2,7,... and the next higher level exists, consider
        // transitions from and to higher level 
        if( ( (l % 5) == 2) && n > l ){
            event = catRoom1.InsControllableEvent("c"+ToStringInteger(l+1)+"-"+lstr+"("+catstr+")" );
            catRoom1.SetTransition(absentState,event,presentState);
            
            event = catRoom1.InsControllableEvent("c"+lstr+"-"+ToStringInteger(l+1)+"("+catstr+")" );
            catRoom1.SetTransition(presentState,event,absentState);                                         
        }
        // if level 3,8,... and , consider
        // transitions from and to lower level 
        if( (l % 5) == 3){
            event = catRoom1.InsControllableEvent("c"+ToStringInteger(l-1)+"-"+lstr+"("+catstr+")" );
            catRoom1.SetTransition(absentState,event,presentState);
            
            event = catRoom1.InsControllableEvent("c"+lstr+"-"+ToStringInteger(l-1)+"("+catstr+")" );
            catRoom1.SetTransition(presentState,event,absentState);                 
        }
        
        // compose current spec with already computed specs
        tmp = rCatSpec;
        cParallel(tmp,catRoom1,rCatSpec);
        // ===========================================
        // Restrictions in ROOM 2 of the current level
        // ===========================================
        // similar considerations as above
        // for one cat, mutual exclusion with all mice in room 2
        catRoom2.Clear();
        absentState = catRoom2.InsState("!SC2("+lstr+","+catstr+")" );
        presentState = catRoom2.InsState("SC2("+lstr+","+catstr+")" );
        catRoom2.SetInitState(absentState);
        catRoom2.SetMarkedState(absentState);  
        for(int j = 1; j<=k; j++){
            event = catRoom2.InsControllableEvent("m0-2("+lstr+","+ToStringInteger(j)+")" );
            catRoom2.SetTransition(absentState,event,absentState);
            // consider the mice that can come from one level up in level 3,8,...
            if( ( (l % 5) == 3) && n > l ){
                event = catRoom2.InsControllableEvent("m"+ToStringInteger(l+1)+"-"+lstr+"("+ToStringInteger(j)+")" );
                catRoom2.SetTransition(absentState,event,absentState);                
            }
            // consider the mice that can come from one level down in level 4,9,...
            if( (l % 5) == 4){
                event = catRoom2.InsControllableEvent("m"+ToStringInteger(l-1)+"-"+lstr+"("+ToStringInteger(j)+")" );
                catRoom2.SetTransition(absentState,event,absentState);                   
            }
        }
        // insert transitions between the absent and the present state of the cat in room 2
        event = catRoom2.InsControllableEvent("c1-2("+lstr+","+catstr+")" );
        catRoom2.SetTransition(absentState,event,presentState);
        
        event = catRoom2.InsControllableEvent("c2-0("+lstr+","+catstr+")" );
        catRoom2.SetTransition(presentState,event,absentState);
        
        // if level 3,8,... and the next higher level exists, consider
        // transitions from and to higher level 
        if( ( (l % 5) == 3) && n > l ){
            event = catRoom2.InsControllableEvent("c"+ToStringInteger(l+1)+"-"+lstr+"("+catstr+")" );
            catRoom2.SetTransition(absentState,event,presentState);
            
            event = catRoom2.InsControllableEvent("c"+lstr+"-"+ToStringInteger(l+1)+"("+catstr+")" );
            catRoom2.SetTransition(presentState,event,absentState);                                         
        }
        // if level 4,9,... and , consider
        // transitions from and to lower level 
        if( (l % 5) == 4){
            event = catRoom2.InsControllableEvent("c"+ToStringInteger(l-1)+"-"+lstr+"("+catstr+")" );
            catRoom2.SetTransition(absentState,event,presentState);
            
            event = catRoom2.InsControllableEvent("c"+lstr+"-"+ToStringInteger(l-1)+"("+catstr+")" );
            catRoom2.SetTransition(presentState,event,absentState);                 
        }
        
        // compose current spec with already computed spec
        tmp = rCatSpec;
        cParallel(tmp,catRoom2,rCatSpec);
        // ===========================================
        // Restrictions in ROOM 3 of the current level
        // ===========================================
        // similar considerations as above
        // for one cat, mutual exclusion with all mice in room 3
        catRoom3.Clear();
        absentState = catRoom3.InsState("!SC3("+lstr+","+catstr+")" );
        presentState = catRoom3.InsState("SC3("+lstr+","+catstr+")" );
        catRoom3.SetInitState(absentState);
        catRoom3.SetMarkedState(absentState);  
        for(int j = 1; j<=k; j++){
            event = catRoom3.InsControllableEvent("m4-3("+lstr+","+ToStringInteger(j)+")" );
            catRoom3.SetTransition(absentState,event,absentState);
            // consider the mice that can come from one level up in level 4,9,...
            if( ( (l % 5) == 4) && n > l ){
                event = catRoom3.InsControllableEvent("m"+ToStringInteger(l+1)+"-"+lstr+"("+ToStringInteger(j)+")" );
                catRoom3.SetTransition(absentState,event,absentState);                
            }
            // consider the mice that can come from one level down in level 5,10,...
            if( (l % 5) == 0){
                event = catRoom3.InsControllableEvent("m"+ToStringInteger(l-1)+"-"+lstr+"("+ToStringInteger(j)+")" );
                catRoom3.SetTransition(absentState,event,absentState);                   
            }
        }
        // insert transitions between the absent and the present state of the cut in room 3
        event = catRoom3.InsControllableEvent("c0-3("+lstr+","+catstr+")" );
        catRoom3.SetTransition(absentState,event,presentState);
        
        event = catRoom3.InsUncontrollableEvent("c1-3("+lstr+","+catstr+")" );
        catRoom3.SetTransition(absentState,event,presentState);
        
        event = catRoom3.InsControllableEvent("c3-4("+lstr+","+catstr+")" );
        catRoom3.SetTransition(presentState,event,absentState);
        
        event = catRoom3.InsUncontrollableEvent("c3-1("+lstr+","+catstr+")" );
        catRoom3.SetTransition(presentState,event,absentState);

        // if level 4,9,... and the next higher level exists, consider
        // transitions from and to higher level 
        if( ( (l % 5) == 4) && n > l ){
            event = catRoom3.InsControllableEvent("c"+ToStringInteger(l+1)+"-"+lstr+"("+catstr+")" );
            catRoom3.SetTransition(absentState,event,presentState);
            
            event = catRoom3.InsControllableEvent("c"+lstr+"-"+ToStringInteger(l+1)+"("+catstr+")" );
            catRoom3.SetTransition(presentState,event,absentState);                                         
        }
        // if level 5,10,... and , consider
        // transitions from and to lower level 
        if( (l % 5) == 0){
            event = catRoom3.InsControllableEvent("c"+ToStringInteger(l-1)+"-"+lstr+"("+catstr+")" );
            catRoom3.SetTransition(absentState,event,presentState);
            
            event = catRoom3.InsControllableEvent("c"+lstr+"-"+ToStringInteger(l-1)+"("+catstr+")" );
            catRoom3.SetTransition(presentState,event,absentState);                 
        }
                    
        // compose current spec with already computed specs
        tmp = rCatSpec;
        cParallel(tmp,catRoom3,rCatSpec);
        // ===========================================
        // Restrictions in ROOM 4 of the current level
        // ===========================================
        // similar considerations as above
        // for one cat, mutual exclusion with all mice in room 4
        catRoom4.Clear();
        absentState = catRoom4.InsState("!SC4("+lstr+","+catstr+")" );
        presentState = catRoom4.InsState("SC4("+lstr+","+catstr+")" );
        catRoom4.SetInitState(absentState);
        catRoom4.SetMarkedState(absentState);  
        for(int j = 1; j<=k; j++){
            event = catRoom4.InsControllableEvent("m0-4("+lstr+","+ToStringInteger(j)+")" );
            catRoom4.SetTransition(absentState,event,absentState);
            // consider the mice that can come from one level up in level 5,10,...
            if( ( (l % 5) == 0) && n > l ){
                event = catRoom4.InsControllableEvent("m"+ToStringInteger(l+1)+"-"+lstr+"("+ToStringInteger(j)+")" );
                catRoom4.SetTransition(absentState,event,absentState);                
            }
            // consider the mice that can come from one level down in level 6,11,...
            if( (l % 5) == 1 && l> 1){
                event = catRoom4.InsControllableEvent("m"+ToStringInteger(l-1)+"-"+lstr+"("+ToStringInteger(j)+")" );
                catRoom4.SetTransition(absentState,event,absentState);                   
            }
        }
        // insert transitions between the absent and the present state of the cat in room 4
        event = catRoom4.InsControllableEvent("c3-4("+lstr+","+catstr+")" );
        catRoom4.SetTransition(absentState,event,presentState);
        
        event = catRoom4.InsControllableEvent("c4-0("+lstr+","+catstr+")" );
        catRoom4.SetTransition(presentState,event,absentState);
        
        // if level 5,10,... and the next higher level exists, consider
        // transitions from and to higher level 
        if( ( (l % 5) == 0) && n > l ){
            event = catRoom4.InsControllableEvent("c"+ToStringInteger(l+1)+"-"+lstr+"("+catstr+")" );
            catRoom4.SetTransition(absentState,event,presentState);
            
            event = catRoom4.InsControllableEvent("c"+lstr+"-"+ToStringInteger(l+1)+"("+catstr+")" );
            catRoom4.SetTransition(presentState,event,absentState);                                         
        }
        // if level 6,11,... and , consider
        // transitions from and to lower level 
        if( (l % 5) == 1 && l > 1){
            event = catRoom4.InsControllableEvent("c"+ToStringInteger(l-1)+"-"+lstr+"("+catstr+")" );
            catRoom4.SetTransition(absentState,event,presentState);
            
            event = catRoom4.InsControllableEvent("c"+lstr+"-"+ToStringInteger(l-1)+"("+catstr+")" );
            catRoom4.SetTransition(presentState,event,absentState);                 
        }
        
        // compose current spec with already computed specs
        tmp = rCatSpec;
        cParallel(tmp,catRoom4,rCatSpec);        
    }// for(int l = 1; l <= n; l++)
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%% Create specification of mouse k (residing on n maze levels) %%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// MouseSpecification(rMouse, rMouseSpec, mouseNumber, n, k)
void MouseSpecification(cGenerator& rMouse, cGenerator& rMouseSpec, const int mouseNumber, const int n, const int k){
    FD_DF("MouseSpecification(" << rMouse.Name() << "," << rMouseSpec.Name() << "," << mouseNumber << "," << n << "," << k << ")");
    Idx absentState, presentState, event;
    std::string kstr = ToStringInteger(k);
    // vectors for all specifications in all rooms in all levels: these specification state that the mouse rMouse
    // must never be in the same room with any of the cats
    std::vector<cGenerator> miceRooms0(k*n);
    cGenerator mouseRoom0;
    std::vector<cGenerator> miceRooms1(k*n);
    cGenerator mouseRoom1;
    std::vector<cGenerator> miceRooms2(k*n);
    cGenerator mouseRoom2;
    std::vector<cGenerator> miceRooms3(k*n);
    cGenerator mouseRoom3;
    std::vector<cGenerator> miceRooms4(k*n);
    cGenerator mouseRoom4;
    // prepare the specification generator
    rMouseSpec.Clear();
    cGenerator tmp;
    // the specification is initialized with the corresponding mouse generator
    Mouse(rMouseSpec,n,mouseNumber);
    rMouse = rMouseSpec;
    std::string mousestr = ToStringInteger(mouseNumber);
    
    // go through all rooms and levels, and determine the restrictions for the mouse mouseNumber
    for(int l = 1; l <= n; l++){  
        std::string lstr = ToStringInteger(l);
        mouseRoom0.Clear();
        // ============================================
        // Restrictions in ROOM 0 for the current level
        // ============================================
        // for mouse mouseNumber, mutual exclusion with all cats in room 0: whenever the mouse is absent, any 
        // cat can be in the room, if the mouse is present, no cat may enter the room
        absentState = mouseRoom0.InsState("!SM0("+lstr+","+mousestr+")" );
        presentState = mouseRoom0.InsState("SM0("+lstr+","+mousestr+")" );
        // the mouse is initially absent in room 0
        mouseRoom0.SetInitState(absentState);
        mouseRoom0.SetMarkedState(absentState);  
        // insert all cat events that are allowed in the absent state of the mouse
        for(int j = 1; j<=k; j++){
            event = mouseRoom0.InsControllableEvent("c2-0("+lstr+","+ToStringInteger(j)+")" );
            mouseRoom0.SetTransition(absentState,event,absentState);
            
            event = mouseRoom0.InsControllableEvent("c4-0("+lstr+","+ToStringInteger(j)+")" );
            mouseRoom0.SetTransition(absentState,event,absentState);
            // consider the cats that can come from one level up in level 1,6,...
            if( ( (l % 5) == 1) && n > l ){
                event = mouseRoom0.InsControllableEvent("c"+ToStringInteger(l+1)+"-"+lstr+"("+ToStringInteger(j)+")" );
                mouseRoom0.SetTransition(absentState,event,absentState);                
            }
            // consider the mice that can come from one level down in level 2,7,...
            if( (l % 5) == 2){
                event = mouseRoom0.InsControllableEvent("c"+ToStringInteger(l-1)+"-"+lstr+"("+ToStringInteger(j)+")" );
                mouseRoom0.SetTransition(absentState,event,absentState);                   
            }
        }
        // insert the transisiont between the absent and the present state of the mouse in room 0
        event = mouseRoom0.InsControllableEvent("m1-0("+lstr+","+mousestr+")" );
        mouseRoom0.SetTransition(absentState,event,presentState);
        
        event = mouseRoom0.InsControllableEvent("m3-0("+lstr+","+mousestr+")" );
        mouseRoom0.SetTransition(absentState,event,presentState);
        
        event = mouseRoom0.InsControllableEvent("m0-2("+lstr+","+mousestr+")" );
        mouseRoom0.SetTransition(presentState,event,absentState);
        
        event = mouseRoom0.InsControllableEvent("m0-4("+lstr+","+mousestr+")" );
        mouseRoom0.SetTransition(presentState,event,absentState);
        // if level 1,6,... and the next higher level exists, consider
        // transitions from and to higher level 
        if( ( (l % 5) == 1) && n > l ){
            event = mouseRoom0.InsControllableEvent("m"+ToStringInteger(l+1)+"-"+lstr+"("+mousestr+")" );
            mouseRoom0.SetTransition(absentState,event,presentState);
            
            event = mouseRoom0.InsControllableEvent("m"+lstr+"-"+ToStringInteger(l+1)+"("+mousestr+")" );
            mouseRoom0.SetTransition(presentState,event,absentState);                                         
        }
        // if level 2,7,... and , consider
        // transitions from and to lower level 
        if( (l % 5) == 2){
            event = mouseRoom0.InsControllableEvent("m"+ToStringInteger(l-1)+"-"+lstr+"("+mousestr+")" );
            mouseRoom0.SetTransition(absentState,event,presentState);
            
            event = mouseRoom0.InsControllableEvent("m"+lstr+"-"+ToStringInteger(l-1)+"("+mousestr+")" );
            mouseRoom0.SetTransition(presentState,event,absentState);                 
        }
        
        // compose current spec with already existing specs
        tmp = rMouseSpec;
        cParallel(tmp,mouseRoom0,rMouseSpec);

        // ===========================================
        // Restrictions in ROOM 1 of the current level
        // ===========================================
        // similar considerations to above
        mouseRoom1.Clear();
        // for one mouse, mutual exclusion with all cats in room 1
        absentState = mouseRoom1.InsState("!SM1("+lstr+","+mousestr+")" );
        presentState = mouseRoom1.InsState("SM1("+lstr+","+mousestr+")" );
        mouseRoom1.SetInitState(absentState);
        mouseRoom1.SetMarkedState(absentState);  
        for(int j = 1; j<=k; j++){
            event = mouseRoom1.InsControllableEvent("c0-1("+lstr+","+ToStringInteger(j)+")" );
            mouseRoom1.SetTransition(absentState,event,absentState);
            
            event = mouseRoom1.InsUncontrollableEvent("c3-1("+lstr+","+ToStringInteger(j)+")" );
            mouseRoom1.SetTransition(absentState,event,absentState);
            // consider the cats that can come from one level up in level 2,7,...
            if( ( (l % 5) == 2) && n > l ){
                event = mouseRoom1.InsControllableEvent("c"+ToStringInteger(l+1)+"-"+lstr+"("+ToStringInteger(j)+")" );
                mouseRoom1.SetTransition(absentState,event,absentState);                
            }
            // consider the mice that can come from one level down in level 3,8,...
            if( (l % 5) == 3){
                event = mouseRoom1.InsControllableEvent("c"+ToStringInteger(l-1)+"-"+lstr+"("+ToStringInteger(j)+")" );
                mouseRoom1.SetTransition(absentState,event,absentState);                   
            }
        }
        // insert all transitions between the absent and the present state of the mouse in room 1
        event = mouseRoom1.InsControllableEvent("m2-1("+lstr+","+mousestr+")" );
        mouseRoom1.SetTransition(absentState,event,presentState);
        
        event = mouseRoom1.InsControllableEvent("m1-0("+lstr+","+mousestr+")" );
        mouseRoom1.SetTransition(presentState,event,absentState);
        // if level 2,7,... and the next higher level exists, consider
        // transitions from and to higher level 
        if( ( (l % 5) == 2) && n > l ){
            event = mouseRoom1.InsControllableEvent("m"+ToStringInteger(l+1)+"-"+lstr+"("+mousestr+")" );
            mouseRoom1.SetTransition(absentState,event,presentState);
            
            event = mouseRoom1.InsControllableEvent("m"+lstr+"-"+ToStringInteger(l+1)+"("+mousestr+")" );
            mouseRoom1.SetTransition(presentState,event,absentState);                                         
        }
        // if level 3,8,... and , consider
        // transitions from and to lower level 
        if( (l % 5) == 3){
            event = mouseRoom1.InsControllableEvent("m"+ToStringInteger(l-1)+"-"+lstr+"("+mousestr+")" );
            mouseRoom1.SetTransition(absentState,event,presentState);
            
            event = mouseRoom1.InsControllableEvent("m"+lstr+"-"+ToStringInteger(l-1)+"("+mousestr+")" );
            mouseRoom1.SetTransition(presentState,event,absentState);                 
        }
        
        // compose current spec with already existing specs
        tmp = rMouseSpec;
        cParallel(tmp,mouseRoom1,rMouseSpec);
        
        // ==========================================
        // Restriction in ROOM 2 of the current level
        // ==========================================
        // for one mouse, mutual exclusion with all cats in room 2
        // similar considerations as above
        mouseRoom2.Clear();
        absentState = mouseRoom2.InsState("!SM0("+lstr+","+mousestr+")" );
        presentState = mouseRoom2.InsState("SM0("+lstr+","+mousestr+")" );
        mouseRoom2.SetInitState(absentState);
        mouseRoom2.SetMarkedState(absentState);  
        for(int j = 1; j<=k; j++){
            event = mouseRoom2.InsControllableEvent("c1-2("+lstr+","+ToStringInteger(j)+")" );
            mouseRoom2.SetTransition(absentState,event,absentState);
            // consider the cats that can come from one level up in level 3,8,...
            if( ( (l % 5) == 3) && n > l ){
                event = mouseRoom2.InsControllableEvent("c"+ToStringInteger(l+1)+"-"+lstr+"("+ToStringInteger(j)+")" );
                mouseRoom2.SetTransition(absentState,event,absentState);                
            }
            // consider the mice that can come from one level down in level 4,9,...
            if( (l % 5) == 4){
                event = mouseRoom2.InsControllableEvent("c"+ToStringInteger(l-1)+"-"+lstr+"("+ToStringInteger(j)+")" );
                mouseRoom2.SetTransition(absentState,event,absentState);                   
            }
        }
        // insert transitions between the absent and the present state of the mouse in room 2
        event = mouseRoom2.InsControllableEvent("m0-2("+lstr+","+mousestr+")" );
        mouseRoom2.SetTransition(absentState,event,presentState);
        
        event = mouseRoom2.InsControllableEvent("m2-1("+lstr+","+mousestr+")" );
        mouseRoom2.SetTransition(presentState,event,absentState);
        // if level 3,8,... and the next higher level exists, consider
        // transitions from and to higher level 
        if( ( (l % 5) == 3) && n > l ){
            event = mouseRoom2.InsControllableEvent("m"+ToStringInteger(l+1)+"-"+lstr+"("+mousestr+")" );
            mouseRoom2.SetTransition(absentState,event,presentState);
            
            event = mouseRoom2.InsControllableEvent("m"+lstr+"-"+ToStringInteger(l+1)+"("+mousestr+")" );
            mouseRoom2.SetTransition(presentState,event,absentState);                                         
        }
        // if level 4,9,... and , consider
        // transitions from and to lower level 
        if( (l % 5) == 4){
            event = mouseRoom2.InsControllableEvent("m"+ToStringInteger(l-1)+"-"+lstr+"("+mousestr+")" );
            mouseRoom2.SetTransition(absentState,event,presentState);
            
            event = mouseRoom2.InsControllableEvent("m"+lstr+"-"+ToStringInteger(l-1)+"("+mousestr+")" );
            mouseRoom2.SetTransition(presentState,event,absentState);                 
        }
        
        // compose current spec with already existing specs
        tmp = rMouseSpec;
        cParallel(tmp,mouseRoom2,rMouseSpec);
        
        // ===========================================
        // Restrictions in ROOM 3 of rhe current level
        // ===========================================
        // for one mouse, mutual exclusion with all cats in room 3
        // similar considerations as above
        mouseRoom3.Clear();
        absentState = mouseRoom3.InsState("!SM3("+lstr+","+mousestr+")" );
        presentState = mouseRoom3.InsState("SM3("+lstr+","+mousestr+")" );
        mouseRoom3.SetInitState(absentState);
        mouseRoom3.SetMarkedState(absentState);  
        for(int j = 1; j<=k; j++){
            event = mouseRoom3.InsControllableEvent("c0-3("+lstr+","+ToStringInteger(j)+")" );
            mouseRoom3.SetTransition(absentState,event,absentState);
            
            event = mouseRoom3.InsUncontrollableEvent("c1-3("+lstr+","+ToStringInteger(j)+")" );
            mouseRoom3.SetTransition(absentState,event,absentState);                
            // consider the cats that can come from one level up in level 4,9,...
            if( ( (l % 5) == 4) && n > l ){
                event = mouseRoom3.InsControllableEvent("c"+ToStringInteger(l+1)+"-"+lstr+"("+ToStringInteger(j)+")" );
                mouseRoom3.SetTransition(absentState,event,absentState);                
            }
            // consider the mice that can come from one level down in level 5,10,...
            if( (l % 5) == 0){
                event = mouseRoom3.InsControllableEvent("c"+ToStringInteger(l-1)+"-"+lstr+"("+ToStringInteger(j)+")" );
                mouseRoom3.SetTransition(absentState,event,absentState);                   
            }
        }
        // insert transitions between the absent and present state of the mouse in room 3
        event = mouseRoom3.InsControllableEvent("m4-3("+lstr+","+mousestr+")" );
        mouseRoom3.SetTransition(absentState,event,presentState);
        
        event = mouseRoom3.InsControllableEvent("m3-0("+lstr+","+mousestr+")" );
        mouseRoom3.SetTransition(presentState,event,absentState);
        // if level 4,9,... and the next higher level exists, consider
        // transitions from and to higher level 
        if( ( (l % 5) == 4) && n > l ){
            event = mouseRoom3.InsControllableEvent("m"+ToStringInteger(l+1)+"-"+lstr+"("+mousestr+")" );
            mouseRoom3.SetTransition(absentState,event,presentState);
            
            event = mouseRoom3.InsControllableEvent("m"+lstr+"-"+ToStringInteger(l+1)+"("+mousestr+")" );
            mouseRoom3.SetTransition(presentState,event,absentState);                                         
        }
        // if level 5,10,... and , consider
        // transitions from and to lower level 
        if( (l % 5) == 0){
            event = mouseRoom3.InsControllableEvent("m"+ToStringInteger(l-1)+"-"+lstr+"("+mousestr+")" );
            mouseRoom3.SetTransition(absentState,event,presentState);
            
            event = mouseRoom3.InsControllableEvent("m"+lstr+"-"+ToStringInteger(l-1)+"("+mousestr+")" );
            mouseRoom3.SetTransition(presentState,event,absentState);                 
        }
        
        // compose current spec with already existing specs
        tmp = rMouseSpec;
        cParallel(tmp,mouseRoom3,rMouseSpec);
        
        // ==========================================
        // Restriction in ROOM 4 of the current level
        // ==========================================
        // for one mouse, mutual exclusion with all cats in room 4
        // similar considerations as above
        mouseRoom4.Clear();
        absentState = mouseRoom4.InsState("!SM4("+lstr+","+mousestr+")" );
        presentState = mouseRoom4.InsState("SM4("+lstr+","+mousestr+")" );
        if(l == n){
            mouseRoom4.SetInitState(presentState);
            mouseRoom4.SetMarkedState(presentState);  
        }
        else{
             mouseRoom4.SetInitState(absentState);
            mouseRoom4.SetMarkedState(absentState);               
        }
        for(int j = 1; j<=k; j++){
            event = mouseRoom4.InsControllableEvent("c3-4("+lstr+","+ToStringInteger(j)+")" );
            mouseRoom4.SetTransition(absentState,event,absentState);
            // consider the cats that can come from one level up in level 5,10,...
            if( ( (l % 5) == 0) && n > l ){
                event = mouseRoom4.InsControllableEvent("c"+ToStringInteger(l+1)+"-"+lstr+"("+ToStringInteger(j)+")" );
                mouseRoom4.SetTransition(absentState,event,absentState);                
            }
            // consider the mice that can come from one level down in level 6,11,...
            if( (l % 5) == 1 && l>1){
                event = mouseRoom4.InsControllableEvent("c"+ToStringInteger(l-1)+"-"+lstr+"("+ToStringInteger(j)+")" );
                mouseRoom4.SetTransition(absentState,event,absentState);                   
            }
        }
        // insert transitions between the absent and the present state of the mouse in room 4
        event = mouseRoom4.InsControllableEvent("m0-4("+lstr+","+mousestr+")" );
        mouseRoom4.SetTransition(absentState,event,presentState);
        
        event = mouseRoom4.InsControllableEvent("m4-3("+lstr+","+mousestr+")" );
        mouseRoom4.SetTransition(presentState,event,absentState);
        // if level 5,10,... and the next higher level exists, consider
        // transitions from and to higher level 
        if( ( (l % 5) == 0) && n > l ){
            event = mouseRoom4.InsControllableEvent("m"+ToStringInteger(l+1)+"-"+lstr+"("+mousestr+")" );
            mouseRoom4.SetTransition(absentState,event,presentState);
            
            event = mouseRoom4.InsControllableEvent("m"+lstr+"-"+ToStringInteger(l+1)+"("+mousestr+")" );
            mouseRoom4.SetTransition(presentState,event,absentState);                                         
        }
        // if level 6,11,... and , consider
        // transitions from and to lower level 
        if( (l % 5) == 1 && l > 1){
            event = mouseRoom4.InsControllableEvent("m"+ToStringInteger(l-1)+"-"+lstr+"("+mousestr+")" );
            mouseRoom4.SetTransition(absentState,event,presentState);
            
            event = mouseRoom4.InsControllableEvent("m"+lstr+"-"+ToStringInteger(l-1)+"("+mousestr+")" );
            mouseRoom4.SetTransition(presentState,event,absentState);                 
        } 
        // compose current spec with already existing specs
        tmp = rMouseSpec;
        cParallel(tmp,mouseRoom4,rMouseSpec);     
    }
}

