/** @file attributes.cpp Classes AttributeVoid and AttributeFlags  */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "attributes.h"


namespace faudes {

/***********************************************************************************
 *
 * implementation of AttributeVoid 
 *
 */

//Write(rTr)
void AttributeVoid::Write(TokenWriter& rTw) const {
}

//Read(rTr)
void AttributeVoid::Read(TokenReader& rTr) {
  Token token;
  while(rTr.Peek(token)) {
    // break on index or name, since this is the next element
    if((token.Type()==Token::String) || (token.Type()==Token::Integer)) {
      break;
    }
    // break on end, since this is the end of the set
    if(token.Type()==Token::End) {
      break;
    }
    // break on Consecutive section, since this belongs to the set
    if((token.Type()==Token::Begin) && (token.StringValue() == "Consecutive")) {
      break;
    }
    // skip any attribute section from other file format
    if(token.Type()==Token::Begin){
      rTr.ReadBegin(token.StringValue());
      rTr.ReadEnd(token.StringValue());
      continue;
    }  
    // skip any other token from other file format
    rTr.Get(token);
  }
}



/***********************************************************************************
 *
 * implementation of AttributeFlags 
 *
 */

//Write(rTw)
// Note: you should write attributes in a section, so that
// the AttributeVoid read method can detect and skip them.
// Here, we make an execption of the rule ...
void AttributeFlags::Write(TokenWriter& rTw) const {
  if(!IsDefault()) {
    Token token;
    token.SetInteger16(mFlags);
    rTw << token;
  }
}

//Write()
void AttributeFlags::Write(void) const {
  TokenWriter rTw(TokenWriter::Stdout);
  Write(rTw);
}


//Read(rTr)
void AttributeFlags::Read(TokenReader& rTr) {
  Token token;
  rTr.Peek(token);
  if(token.Type()==Token::Integer16) { 
    rTr.Get(token);
    mFlags=token.IntegerValue();
  } else {
    mFlags=mDefFlags;
  }
}


} // namespace
