/* luafaudes_func.cpp - Lua bindings for the FAUDES functions */

/* luafaudes - FAUDES Bindings for Lua

Copyright (C) 2006 Bernd Opitz
Exclusive copyright is granted to the Lehrstuhl fuer Regelungstechnik
at the University of Erlangen-Nuernberg, Germany

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. */


#include "luafaudes_func.hpp"

/////////////////////////
// Generator functions:
/////////////////////////


int Accessible(lua_State* L) {
  fGenerator* g = luaL_checkgenerator(L, 1);
  if (g->Accessible()) {
    lua_pushboolean(L, 1);
  }
  else {
    lua_pushboolean(L, 0);
  }
  return 1;
}

int IsAccessible(lua_State* L) {
  fGenerator* g = luaL_checkgenerator(L, 1);
  if (g->IsAccessible()) {
    lua_pushboolean(L, 1);
  }
  else {
    lua_pushboolean(L, 0);
  }
  return 1;
}

int Coaccessible(lua_State* L) {
  lua_checkargs(L,1,1);
  fGenerator* g = luaL_checkgenerator(L, 1);
  if (g->Coaccessible()) {
    lua_pushboolean(L, 1);
  }
  else {
    lua_pushboolean(L, 0);
  }
  return 1;
}

int IsCoaccessible(lua_State* L) {
  lua_checkargs(L,1,1);
  fGenerator* g = luaL_checkgenerator(L, 1);
  if (g->IsCoaccessible()) {
    lua_pushboolean(L, 1);
  }
  else {
    lua_pushboolean(L, 0);
  }
  return 1;
}

int Trim(lua_State* L) {
  lua_checkargs(L,1,1);
  fGenerator* g = luaL_checkgenerator(L, 1);
  if (g->Trim()) {
    lua_pushboolean(L, 1);
  }
  else {
    lua_pushboolean(L, 0);
  }
  return 1;
}

int IsTrim(lua_State* L) {
  lua_checkargs(L,1,1);
  fGenerator* g = luaL_checkgenerator(L, 1);
  if (g->IsTrim()) {
    lua_pushboolean(L, 1);
  }
  else {
    lua_pushboolean(L, 0);
  }
  return 1;
}

// todo: make alphabet optional
int IsControllable(lua_State* L) {
  lua_checkargs(L,3,3);
  fGenerator* g_plant = luaL_checkgenerator(L, 1);
  fEventSet* a_cont = luaL_checkeventset(L, 2);
  fGenerator* g_spec = luaL_checkgenerator(L, 3);
  if (IsControllable(*g_plant, *a_cont, *g_spec)) {
    lua_pushboolean(L, 1);
  }
  else {
    lua_pushboolean(L, 0);
  }
  return 1;
}

int IsControllableResultToString(lua_State* L) {
  lua_checkargs(L,3,3);
  fGenerator* g_plant = luaL_checkgenerator(L, 1);
  fEventSet* a_cont = luaL_checkeventset(L, 2);
  fGenerator* g_spec = luaL_checkgenerator(L, 3);
  StateSet forbidden;
  bool controllable = IsControllable(*g_plant, *a_cont, *g_spec, forbidden);
  std::string str = ConsoleText("Controllability Result", forbidden.ToString());
  lua_pushlstring(L, str.c_str(), str.size());
  if (controllable) {
    lua_pushboolean(L, 1);
  }
  else {
    lua_pushboolean(L, 0);
  }
  return 2;
}

int Deterministic(lua_State* L) {
  lua_checkargs(L,2,2);
  fGenerator* g_nondet = luaL_checkgenerator(L, 1);
  fGenerator* g_det = luaL_checkgenerator(L, 2);
  Deterministic(*g_nondet, *g_det);
  return 0;
}

int InvProject(lua_State* L) {
  lua_checkargs(L,2,2);
  fGenerator* g = luaL_checkgenerator(L, 1);
  fEventSet* invalph = luaL_checkeventset(L, 2);
  InvProject(*g, *invalph);
  return 0;
}

int IsDeterministic(lua_State* L) {
  lua_checkargs(L,1,1);
  fGenerator* g = luaL_checkgenerator(L, 1);
  if (g->IsDeterministic()) {
    lua_pushboolean(L, 1);
  }
  else {
    lua_pushboolean(L, 0);
  }
  return 1;
}

int LanguageComplement(lua_State* L) {
  lua_checkargs(L,1,1);
  fGenerator* g = luaL_checkgenerator(L, 1);
  LanguageComplement(*g);
  return 0;
}

int LanguageConcatenate(lua_State* L) {
  lua_checkargs(L,3,3);
  fGenerator* g_in1 = luaL_checkgenerator(L, 1);
  fGenerator* g_in2 = luaL_checkgenerator(L, 2);
  fGenerator* g_out = luaL_checkgenerator(L, 3);
  LanguageConcatenate(*g_in1, *g_in2, *g_out);
  return 0;
}

int LanguageConcatenateNonDet(lua_State* L) {
  lua_checkargs(L,3,3);
  fGenerator* g_in1 = luaL_checkgenerator(L, 1);
  fGenerator* g_in2 = luaL_checkgenerator(L, 2);
  fGenerator* g_out = luaL_checkgenerator(L, 3);
  LanguageConcatenateNonDet(*g_in1, *g_in2, *g_out);
  return 0;
}

int LanguageEquality(lua_State* L) {
  lua_checkargs(L,2,2);
  fGenerator* g_in1 = luaL_checkgenerator(L, 1);
  fGenerator* g_in2 = luaL_checkgenerator(L, 2);
  if (LanguageEquality(*g_in1, *g_in2)) {
    lua_pushboolean(L, 1);
  }
  else {
    lua_pushboolean(L, 2);
  }
  return 1;
}

int LanguageInclusion(lua_State* L) {
  lua_checkargs(L,2,2);
  fGenerator* g_in1 = luaL_checkgenerator(L, 1);
  fGenerator* g_in2 = luaL_checkgenerator(L, 2);
  if (LanguageInclusion(*g_in1, *g_in2)) {
    lua_pushboolean(L, 1);
  }
  else {
    lua_pushboolean(L, 2);
  }
  return 1;
}

int LanguageIntersection(lua_State* L) {
  lua_checkargs(L,3,3);
  fGenerator* g_in1 = luaL_checkgenerator(L, 1);
  fGenerator* g_in2 = luaL_checkgenerator(L, 2);
  fGenerator* g_out = luaL_checkgenerator(L, 3);
  LanguageIntersection(*g_in1, *g_in2, *g_out);
  return 0;
}


int LanguageUnion(lua_State* L) {
  lua_checkargs(L,3,3);
  fGenerator* g1 = luaL_checkgenerator(L, 1);
  fGenerator* g2 = luaL_checkgenerator(L, 2);
  fGenerator* g_union = luaL_checkgenerator(L, 3);
  LanguageUnion(*g1, *g2, *g_union);
  return 0;
}

int LanguageUnionNonDet(lua_State* L) {
  lua_checkargs(L,3,3);
  fGenerator* g1 = luaL_checkgenerator(L, 1);
  fGenerator* g2 = luaL_checkgenerator(L, 2);
  fGenerator* g_union = luaL_checkgenerator(L, 3);
  LanguageUnionNonDet(*g1, *g2, *g_union);
  return 0;
}


int FullLanguage(lua_State* L) {
  lua_checkargs(L,2,2);
  fEventSet* e1 = luaL_checkeventset(L, 1);
  fGenerator* g_out = luaL_checkgenerator(L, 2);
  FullLanguage(*e1, *g_out);
  return 0;
}


int Parallel(lua_State* L) {
  lua_checkargs(L,3,3);
  fGenerator* g_in1 = luaL_checkgenerator(L, 1);
  fGenerator* g_in2 = luaL_checkgenerator(L, 2);
  fGenerator* g_out = luaL_checkgenerator(L, 3);
  std::map< std::pair<Idx,Idx>, Idx> rcmap;
  cParallel(*g_in1, *g_in2, *g_out);
  return 0;
}

int ProjectNonDet(lua_State* L) {
  lua_checkargs(L,2,2);
  fGenerator* g = luaL_checkgenerator(L, 1);
  fEventSet* invalph = luaL_checkeventset(L, 2);
  ProjectNonDet(*g, *invalph);
  return 0;
}

int Project(lua_State* L) {
  lua_checkargs(L,3,3);
  fGenerator* g_orig = luaL_checkgenerator(L, 1);
  fEventSet* hialph = luaL_checkeventset(L, 2);
  fGenerator* g_proj = luaL_checkgenerator(L, 3);
  faudes::Project(*g_orig, *hialph, *g_proj);
  return 0;
}


int StateMin(lua_State* L) {
  lua_checkargs(L,2,2);
  fGenerator* g_in = luaL_checkgenerator(L, 1);
  fGenerator* g_out = luaL_checkgenerator(L, 2);
  StateMin(*g_in, *g_out);
  return 0;
}

int SupCon(lua_State* L) {
  lua_checkargs(L,3,3);
  fGenerator* g_in1 = luaL_checkgenerator(L, 1);
  fGenerator* g_in2 = luaL_checkgenerator(L, 2);
  fGenerator* g_out = luaL_checkgenerator(L, 3);
  cSupCon(*g_in1, *g_in2, *g_out);
  return 0;
}

int SupConNB(lua_State* L) {
  lua_checkargs(L,3,3);
  fGenerator* g_in1 = luaL_checkgenerator(L, 1);
  fGenerator* g_in2 = luaL_checkgenerator(L, 2);
  fGenerator* g_out = luaL_checkgenerator(L, 3);
  cSupConNB(*g_in1, *g_in2, *g_out);
  return 0;
}


int EventSetUnion(lua_State* L) {
  lua_checkargs(L,3,3);
  fEventSet* e1 = luaL_checkeventset(L, 1);
  fEventSet* e2 = luaL_checkeventset(L, 2);
  fEventSet* e_union = luaL_checkeventset(L, 3);
  *e_union = *e1;
  e_union->InsertSet(*e2);
  return 0;
}

int EventSetIntersection(lua_State* L) {
  lua_checkargs(L,3,3);
  fEventSet* e1 = luaL_checkeventset(L, 1);
  fEventSet* e2 = luaL_checkeventset(L, 2);
  fEventSet* e_inter = luaL_checkeventset(L, 3);
  // maintain atributes: need faudes fnct
  fEventSet e2minus1 = *e2;
  e2minus1.EraseSet(*e1);
  *e_inter = *e2;
  e_inter->EraseSet(e2minus1);
  return 0;
}

int EventSetDifference(lua_State* L) {
  lua_checkargs(L,3,3);
  fEventSet* e1 = luaL_checkeventset(L, 1);
  fEventSet* e2 = luaL_checkeventset(L, 2);
  fEventSet* e_diff = luaL_checkeventset(L, 3);
  *e_diff = *e1;
  e_diff->EraseSet(*e2);
  return 0;
}


int HelpFunctions(lua_State* L) {
  std::stringstream sstr;
  sstr
    << std::endl
    << "  boolean Accessible(Generator)" << std::endl
    << "  boolean Coaccessible(Generator)" << std::endl
    << "  boolean IsControllable(Generator, EventSet, Generator)" << std::endl
    << "      nil Deterministic(Generator, Generator)" << std::endl
    << "      nil InvProject(Generator, EventSet)" << std::endl
    << "  boolean IsAccessible(Generator)" << std::endl
    << "  boolean IsCoaccessible(Generator)" << std::endl
    << "  boolean IsTrim(Generator)" << std::endl
    << "  boolean IsDeterministic(Generator)" << std::endl
    << "      nil LanguageComplement(Generator)" << std::endl
    << "      nil LanguageConcatenate(Generator, Generator, Generator)" << std::endl
    << "      nil LanguageConcatenateNonDet(Generator, Generator, Generator)" << std::endl
    << "  boolean LanguageEquality(Generator, Generator)" << std::endl
    << "  boolean LanguageInclusion(Generator, Generator)" << std::endl
    << "      nil LanguageIntersection(Generator, Generator, Generator)" << std::endl
    << "      nil LanguageUnion(Generator, Generator, Generator)" << std::endl
    << "      nil LanguageUnionNonDet(Generator, Generator, Generator)" << std::endl
    << "      nil FullLanguage(EventSet, Generator)" << std::endl
    << "      nil Parallel(Generator, Generator, Generator)" << std::endl
    << "      nil Project(Generator, EventSet, Generator)" << std::endl
    << "      nil ProjectNonDet(Generator, EventSet)" << std::endl
    << "      nil StateMin(Generator, Generator)" << std::endl
    << "      nil SupCon(Generator, Generator, Generator)" << std::endl
    << "      nil SupConNB(Generator, Generator, Generator)" << std::endl
    << "  boolean Trim(Generator)" << std::endl
    << "      nil EventSetUnion(EventSet, EventSet, EventSet)" << std::endl
    << "      nil EventSetIntersection(EventSet, EventSet, EventSet)" << std::endl
    << "      nil EventSetDifference(EventSet, EventSet, EventSet)" << std::endl;
  DisplayText("faudes Functions", sstr.str());
  return 0;
}





///////////////////////////////////////////////
// register functions for Generator data type
///////////////////////////////////////////////


void RegisterGeneratorFunctions(lua_State* L) {
  // register functions:
  lua_register(L, "Accessible", Accessible);
  lua_register(L, "Coaccessible", Coaccessible);
  lua_register(L, "IsControllable", IsControllable);
  lua_register(L, "Deterministic", Deterministic);
  lua_register(L, "InvProject", InvProject);
  lua_register(L, "IsAccessible", IsAccessible);
  lua_register(L, "IsCoaccessible", IsCoaccessible);
  lua_register(L, "IsTrim", IsTrim);
  lua_register(L, "IsDeterministic", IsDeterministic);
  lua_register(L, "LanguageComplement", LanguageComplement);
  lua_register(L, "LanguageConcatenate", LanguageConcatenate);
  lua_register(L, "LanguageConcatenateNonDet", LanguageConcatenateNonDet);
  lua_register(L, "LanguageEquality", LanguageEquality);
  lua_register(L, "LanguageInclusion", LanguageInclusion);
  lua_register(L, "LanguageIntersection", LanguageIntersection);
  lua_register(L, "LanguageUnion", LanguageUnion);
  lua_register(L, "LanguageUnionNonDet", LanguageUnionNonDet);
  lua_register(L, "FullLanguage", FullLanguage);
  lua_register(L, "Parallel", Parallel);
  lua_register(L, "Project", Project);
  lua_register(L, "ProjectNonDet", ProjectNonDet);
  lua_register(L, "StateMin", StateMin);
  lua_register(L, "SupCon", SupCon);
  lua_register(L, "SupConNB", SupConNB);
  lua_register(L, "Trim", Trim);
  lua_register(L, "EventSetUnion", EventSetUnion);
  lua_register(L, "EventSetIntersection", EventSetIntersection);
  lua_register(L, "EventSetDifference", EventSetDifference);
  lua_register(L, "HelpFunctions", HelpFunctions);
}

