-- 
-- luafaudes demo
--
-- Enter "../bin/luafaudes demo.lua" at the commandline
-- to execute this script. It reads data from the directory
-- "data" and writes results with prefix "tmp_"
-- 


-- ------------------------------------------------------------------
-- demo 1: accessibility --------------------------------------------
-- ------------------------------------------------------------------

print('**** demo 1:  accessible/coaccessible/trim ****')

g_notacc=Generator.New('data/g_notacc.gen')     -- read generator file
if not IsAccessible(g_notacc) then              -- test accessibility
  print('g_notacc is not accessible')           -- report
  print('removing non-accessible states')       -- report
  g_acc=Generator.New(g_notacc)                 -- copy generator
  Accessible(g_acc)                             -- make accessible
  g_acc:Write('tmp_g_acc.gen')                  -- write result
end

g_notcoacc=Generator.New('data/g_notcoacc.gen') -- read generator file
if not IsCoaccessible(g_notcoacc) then          -- test coaccessibility
  print('g_notcoacc is not coaccessible')       -- report
  print('removing non-coaccessible states')     -- report
  g_coacc=Generator.New(g_notcoacc)             -- copy generator
  Coaccessible(g_coacc)                         -- make accessible
  g_coacc:Write('tmp_g_coacc.gen')              -- write result
end

g_nottrim=Generator.New('data/g_nottrim.gen')   -- read generator file
if not IsTrim(g_nottrim) then                   -- test trimness
  print('g_nottrim is not trim')                -- report
  print('removing non-(co-)accessible states')  -- report
  g_trim=Generator.New(g_nottrim)               -- copy generator
  Trim(g_trim)                                  -- make trim
  g_trim:Write('tmp_g_trim.gen')                -- write result
end


print('running dot to generate graph output')   -- announce graph output
g_nottrim:GraphWrite('tmp_g_nottrim.png')       -- generate graph output
g_trim:GraphWrite('tmp_g_trim.jpg')             -- generate graph output

-- ------------------------------------------------------------------
-- demo 2: deterministic/statemin  ----------------------------------
-- ------------------------------------------------------------------

print('**** demo 2:  deterministic/statemin ****')

g_nondet=Generator.New('data/g_nondet.gen')     -- read generator file
if not IsDeterministic(g_nondet) then           -- test determinism
  print('g_nondet is non-deterministic')        -- report
  print('convert to deterministic generator')   -- report
  g_det=Generator.New()                         -- new generator
  Deterministic(g_nondet, g_det)                -- make deterministic
  g_det:Write('tmp_g_det.gen')                  -- write result
end

g_notstatemin=Generator.New('data/g_notstatemin.gen')   -- read generator file
print('g_notstatemin may not be minimal')       -- report
print('convert to minimal realisation')         -- report
g_statemin=Generator.New()                      -- new generator
StateMin(g_notstatemin, g_statemin)             -- make stateminerministic
g_statemin:Write('tmp_g_statemin.gen')          -- write result

-- ------------------------------------------------------------------
-- demo 3: (inverse-) projection ------------------------------------
-- ------------------------------------------------------------------

print('**** demo 3:  (inverse-)projection  ****')

gp=Generator.New('data/gp.gen')                 -- read generator file
sig_f=EventSet.New( gp:Alphabet() )             -- copy original alphabet
sig_0=EventSet.New('data/projalph.alp')         -- read alphabet to project on
print('projecting gp')                          -- report
g_proj=Generator.New()                          -- new generator
Project(gp,sig_0,g_proj)                        -- project respective language
print('inverse projection of gp on original alphabet')        -- report
g_invproj=Generator.New(g_proj)                 -- copy generator
InvProject(g_invproj,sig_f)                     -- inverse project respective language
g_proj:Write('tmp_g_proj.gen')                  -- write result
g_invproj:Write('tmp_g_invproj.gen')            -- write result

if LanguageInclusion(g_invproj,gp) then
  print('result includes gp')
  if not LanguageInclusion(gp, g_invproj) then
    print('result strictly includes gp')
  end
end
