------------------------------------------------------------------------------
-- generate_philosopher.lua --
--   create generator modelling philosopher number npar with kpar intermediate
--   states for the WODES'08 "dining philosophers" benchmark example.
--
-- parameters: npar - index of philosopher
--             kpar - number of intermediate states
-- 
-- return: generator model of philosopher
------------------------------------------------------------------------------

-- *** FUNCTION Philosopher: returns philosopher number npar with kpar intermediate states
function Philosopher(npar,kpar)
    
    phil=Generator.New()
    
    phil:Name(string.format('Philosopher_n=%d_k=%d',npar,kpar))
    
    -- create init state nI
    init=phil:InsInitState(string.format('%dI',npar));
    phil:SetMarkedState(init);
    
    -- create transition "Phil n picks left fork" (nfl) to
    -- first of k intermediate states, where n is controllable
    -- if n is odd.
    if((npar % 2)==0) -- if n is even, set nfl uncontrollable
      then
      event = phil:InsUncontrollableEvent(string.format('%dfl',npar))
      else
      event = phil:InsControllableEvent(string.format('%dfl',npar));
    end
    -- create intermediate state 1
    nk1=phil:InsState(string.format('%d_1',npar))
    -- set transition nfl
    phil:SetTransition(init,event,nk1);
    
    -- Create remaining k-1 intermediate states "n_2..n_k"
    -- connected by k-1 intermediate transitions "nt1..ntk-1".
    -- Note: this loop is ignored if k==0 (no intermediate events)
    for i=1,kpar-1 do
      -- insert state n_i+1
      nk2=phil:InsState(string.format('%d_%d',npar,i+1))
      -- insert event nti
      event=phil:InsControllableEvent(string.format('%dt%d',npar,i))
      -- set transition from state n_i to state n_i+1
      phil:SetTransition(nk1,event,nk2);
      -- prepare next iteration
      nk1=nk2
    end
    
    -- set transition "philosopher n takes right fork" (nfr)
    -- from state n_k to eating state nE, and transition
    -- "philosopher n drops both forks" (nf) from state nE to
    -- initial state nI
    nk2=phil:InsState(string.format('%dE',npar))
    event=phil:InsControllableEvent(string.format('%dfr',npar))
    phil:SetTransition(nk1,event,nk2)
    event=phil:InsControllableEvent(string.format('%df',npar))
    phil:SetTransition(nk2,event,init)
    
    return phil
end
-- *** ENDOF FUNCTION philosopher